from __future__ import annotations

from functools import cached_property

from plexfuse.vfs.entry.AttrEntry import AttrEntry


class ControlEntry(AttrEntry):
    def __init__(self, name: str, action):
        self.name = name
        self.action = action

    def read(self, offset: int, size: int):
        content = "\n".join(self.action()).encode() + b"\n"
        return content[offset:offset + size]

    @cached_property
    def size(self):
        # On macOS can't read from 0 byte file
        return 64 * 1024

    def __len__(self):
        return self.size



from __future__ import annotations


class DelayedPropertyCacheControl:
    """
    Cache control methods for class methods, but delayed
    """

    def __init__(self, obj, property: str, factory):
        self.obj = obj
        self.property = property
        self.factory = factory
        self.data = obj.__dict__

    def find_objects(self):
        return (self.factory(s) for s in self.data.get(self.property, []))

    def cache_info(self):
        for c in self.find_objects():
            yield from c.cache_info()

    def cache_clear(self):
        for c in self.find_objects():
            yield from c.cache_clear()



from __future__ import annotations

from typing import TYPE_CHECKING

from plexfuse.normalize import normalize

if TYPE_CHECKING:
    from plexapi.video import Season


class SeasonEntry:
    def __init__(self, item: Season):
        self.item = item
        self.title = normalize(item.title)

    def __str__(self):
        return self.title



from plexfuse.plex.PlexApi import PlexApi
from plexfuse.vfs.ChunkedFile import ChunkedFile


def test_file():
    plex = PlexApi()
    cf = ChunkedFile(plex)
    cf.chunk_size = 16

    key = "/library/parts/38908/1705801932/file.avi"
    # A test reading first chunk
    assert len(cf.read(key, 0, cf.chunk_size)) == cf.chunk_size
    # A test reading chunk that spans two chunk
    assert len(cf.read(key, int(cf.chunk_size / 2), cf.chunk_size)) == cf.chunk_size

    # A test that chunked content is full content
    cf.chunk_size = 6
    offset = 8
    size = 8
    chunked = cf.read(key, offset, size)
    cf.chunk_size = 1024
    unchunked = cf.read(key, offset, size)
    assert chunked == unchunked



from __future__ import annotations

from collections import UserDict, defaultdict
from contextlib import contextmanager
from typing import BinaryIO


class FileCache(UserDict[str, BinaryIO]):
    def __init__(self):
        super().__init__()
        self.nopen = defaultdict(int)

    @contextmanager
    def cached_fh(self, path: str):
        try:
            yield self.open(path)
        finally:
            self.release(path)

    def open(self, path: str):
        if path not in self:
            self[path] = self._open(path)
            self.nopen[path] += 1

        return self[path]

    def release(self, path: str):
        if path not in self:
            raise KeyError

        self.nopen[path] -= 1
        if self.nopen[path] <= 0:
            fh = self[path]
            fh.close()
            del self[path]

    def _open(self, path: str):
        try:
            fh = open(path, "rb")
        except Exception as e:
            raise Exception(e)
        return fh



from __future__ import annotations

from pathlib import Path

from plexfuse.vfs.entry.AttrEntry import AttrEntry


class PathEntry(AttrEntry):
    def __init__(self, path: Path):
        self.path = path
        self.size = path.stat().st_size

    def read(self, offset: int, size: int):
        with self.path.open("rb") as fp:
            fp.seek(offset)
            return fp.read(size)

    def __len__(self):
        return self.size



from __future__ import annotations

from functools import cached_property

from plexapi.exceptions import NotFound

from plexfuse.plex.PlexApi import PlexApi
from plexfuse.vfs.entry.AttrEntry import AttrEntry
from plexfuse.vfs.Playable import Playable


class SubtitleEntry(AttrEntry):
    def __init__(self, playable: Playable, name: str, plex: PlexApi):
        self.playable = playable
        self.name = name
        self.plex = plex

    def read(self, offset: int, size: int):
        with self.path.open("rb") as fp:
            fp.seek(offset)
            return fp.read(size)

    @property
    def attr(self):
        return self.playable.timestamps

    @cached_property
    def path(self):
        try:
            stream = self.playable.subtitles[self.name]
        except KeyError:
            return None

        cache_path = self.plex.cache_path(stream.key)
        if not cache_path.exists():
            print(f"Downloading: {cache_path}")
            try:
                self.plex.download_part(stream.key, cache_path)
            except NotFound as e:
                print(f"ERROR: Subtitle not found: {e}")

        return cache_path

    @cached_property
    def size(self):
        return self.path.stat().st_size

    def __len__(self):
        return self.size



from time import time


class Timestampable:
    def __init__(self, **kw):
        if self.st_ctime == 0:
            self.st_ctime = int(time())
            self.st_atime = self.st_ctime
            self.st_mtime = self.st_ctime



import logging

from plexfuse.plex.PlexApi import PlexApi
from plexfuse.vfs.entry.MovieEntry import MovieEntry

logger = logging.getLogger(__name__)


def test_file():
    plex = PlexApi()
    path = "/library/metadata/1"
    filename = "1.webm"

    item = plex.fetch_item(path)
    media = MovieEntry(item)
    part = plex.media_parts_by_name(media, filename)

    def sizes(gen):
        return list(map(lambda x: len(x), gen))

    assert [1024] == sizes(plex.request_file(part.key, offset=0, size=1024))
    assert [1024] == sizes(plex.request_file(part.key, offset=1024, size=1024))
    assert [512] == sizes(plex.request_file(part.key, offset=1024, size=512))



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections import UserDict


class UserDictCacheControl:
    """
    Cache control methods for an UserDict
    """

    def __init__(self, obj: UserDict):
        self.data = obj.data
        self.name = obj.__class__.__name__

    def cache_info(self):
        yield f"{self.name} cache:"
        for k, v in self.data.items():
            yield f"- {k}: {v.__class__.__name__}"

    def cache_clear(self):
        size = len(self.data)
        self.data.clear()
        yield f"Cleared {self.name} cache: {size} entries"



import unicodedata


def normalize(s: str, is_path=False):
    if not is_path:
        # Handle directory separator in filename
        s = s.replace("/", "∕")
    # Normalize Unicode
    # https://stackoverflow.com/questions/16467479/normalizing-unicode
    return unicodedata.normalize("NFC", s)



from __future__ import annotations

from typing import TYPE_CHECKING

from plexfuse.cache.FileCache import FileCache

if TYPE_CHECKING:
    from plexfuse.plex.PlexApi import PlexApi


class ChunkedFile:
    def __init__(self, plex: PlexApi):
        super().__init__()
        self.plex = plex
        self.chunk_size = plex.CHUNK_SIZE
        self.files = FileCache()

    def read(self, path: str, offset: int, size: int, max_size: int):
        # To avoid reading beyond end of file, adjusting size
        if offset + size > max_size:
            size = max_size - offset
        chunk_number = self.chunk_number(offset)
        chunk_offset = self.chunk_offset(offset)

        reads = []
        while size > 0:
            cache_path = self.cache_path(path, chunk_number)

            with self.files.cached_fh(cache_path) as fp:
                fp.seek(chunk_offset)
                buffer = fp.read(size)
                read_bytes = len(buffer)
                size -= read_bytes
                reads.append(buffer)

            chunk_offset = 0
            chunk_number += 1

        return b"".join(reads)

    def cache_path(self, path: str, chunk_number: int):
        file = self.plex.cache_path(path)
        filename = f"{file.stem}-{self.chunk_size}-{chunk_number}{file.suffix}"
        cache_path = file.parent / filename

        if not cache_path.exists():
            print(f"Downloading: {cache_path}")
            base_offset = self.base_offset(chunk_number)
            self.plex.download_part(path, cache_path,
                                    offset=base_offset, size=self.chunk_size)

        return cache_path

    def base_offset(self, chunk_number: int):
        """ return absolute offset where chunk numner starts """
        return self.chunk_size * chunk_number

    def chunk_number(self, offset: int):
        """ return chunk number where the absolute offset would be """
        return offset // self.chunk_size

    def chunk_offset(self, offset: int):
        """ return relative offset in chunk for absolute offset """
        chunk_number = self.chunk_number(offset)
        base_offset = self.base_offset(chunk_number)
        return offset - base_offset



from collections import UserDict, defaultdict


class RefCountedDict(UserDict):
    def __init__(self):
        super().__init__()
        self.refcount = defaultdict(int)

    def __setitem__(self, key, value):
        if self.refcount[key] > 0 and self.__getitem__(key) != value:
            raise ValueError(f"Value {value} already exists for {key}")

        self.refcount[key] += 1
        super().__setitem__(key, value)

    def __delitem__(self, key):
        self.refcount[key] -= 1
        if self.refcount[key] <= 0:
            super().__delitem__(key)



from __future__ import annotations

from collections import UserDict
from typing import TYPE_CHECKING

from plexfuse.vfs.ChunkedFile import ChunkedFile
from plexfuse.vfs.Control import Control
from plexfuse.vfs.entry.DirEntry import DirEntry
from plexfuse.vfs.entry.FileEntry import FileEntry
from plexfuse.vfs.entry.SubtitleEntry import SubtitleEntry
from plexfuse.vfs.entry.PlexMatchEntry import PlexMatchEntry

if TYPE_CHECKING:
    from plexfuse.fs.PlexFS import PlexFS
    from plexfuse.plex.PlexApi import PlexApi


class PlexVFS(UserDict):
    SUBTITLE_EXT = (".srt", ".vtt")

    def __init__(self, plex: PlexApi, plexfs: PlexFS):
        super().__init__()
        self.plex = plex
        self.reader = ChunkedFile(plex)
        self.control = Control(plex, plexfs, self)

    def __missing__(self, path: str):
        entry = self.resolve(path)
        if entry is None:
            raise KeyError(path)

        self[path] = entry

        return entry

    def resolve(self, path: str):
        if path == "/":
            ents = self.plex.section_types
            ents.extend(self.control.root)
            return DirEntry(ents)

        pe = path.split("/")[1:]
        pc = len(pe)

        try:
            resolved = self.control.handle(pc, pe)
            if resolved is not None:
                return resolved
        except KeyError:
            pass

        if pc == 1 and pe[0] in self["/"]:
            return DirEntry(list(self.plex.sections_by_type(pe[0])))
        elif pc == 2:
            entries = self.plex.library_items(pe[1])
            if entries is None:
                raise KeyError(pe)
            return DirEntry(entries)
        elif pc == 3 and pe[0] == "movie":
            mf = self.plex.movie_files(*pe[1:])
            if mf is None:
                raise KeyError(pe)
            return DirEntry(mf)
        elif pc == 3 and pe[0] == "show":
            seasons = self.plex.show_seasons(*pe[1:])
            if seasons is None:
                raise KeyError(pe)
            return DirEntry(seasons)
        elif pc == 4 and pe[0] in ["movie", "show"] and pe[3].endswith(".plexmatch"):
            playable = self.plex.library_item(*pe[1:-1])
            if playable is None:
                raise KeyError(pe)
            return PlexMatchEntry(playable)
        elif ((pc == 4 and pe[0] == "movie") or (pc == 6 and pe[0] == "show")) and pe[-1].endswith(self.SUBTITLE_EXT):
            if pe[0] == "movie":
                playable = self.plex.library_item(*pe[1:-1])
            elif pe[0] == "show":
                playable = self.plex.show_episode(*pe[1:-1])
            else:
                raise KeyError(pe)
            return SubtitleEntry(playable, name=pe[-1], plex=self.plex)
        elif pc == 4 and pe[0] == "show":
            episodes = self.plex.season_episodes(*pe[1:])
            if episodes is None:
                raise KeyError(pe)
            return DirEntry(episodes)
        elif pc == 5 and pe[0] == "show":
            names = self.plex.episode_files(*pe[1:])
            if names is None:
                raise KeyError(pe)
            return DirEntry(names)
        elif pc == 6 and pe[0] == "show":
            packed = self.plex.episode_part(*pe[1:])
            if packed is None:
                raise KeyError(pe)
            part, playable = packed
            if part is None or playable is None:
                raise KeyError(pe)
            return FileEntry(part, playable=playable, reader=self.reader)
        elif pc == 4 and pe[0] == "movie":
            packed = self.plex.movie_part(*pe[1:])
            if packed is None:
                raise KeyError(pe)
            part, playable = packed
            if part is None or playable is None:
                raise KeyError(pe)
            return FileEntry(part, playable=playable, reader=self.reader)

        return None



from __future__ import annotations

from collections import defaultdict
from functools import cached_property
from typing import TYPE_CHECKING

from plexfuse.normalize import normalize
from plexfuse.vfs.entry.SectionEntry import SectionEntry

if TYPE_CHECKING:
    from plexapi.library import Library


class LibraryEntry:
    def __init__(self, library: Library):
        self.library = library

    @cached_property
    def sections(self):
        return [SectionEntry(s) for s in self.library.sections()]

    @property
    def section_types(self):
        return self.sections_by_type.keys()

    @cached_property
    def sections_by_type(self):
        d = defaultdict(list)
        for s in self.sections:
            # Not yet there
            if s.type == "artist":
                continue
            d[s.type].append(s)
        return d

    @cached_property
    def section_by_title(self):
        return dict({(normalize(s.title), s) for s in self.sections})



from plexfuse.fs.RefCountedDict import RefCountedDict


def test_file():
    d = RefCountedDict()

    d["key1"] = "value1"
    assert "key1" in d, "set value and it's there"
    assert d["key1"] == "value1"

    d["key1"] = "value1"
    assert "key1" in d, "set value again with same value increases refcount"
    assert d["key1"] == "value1"

    try:
        d["key1"] = "value2"
        assert False, "setting different value must throw KeyError"
    except ValueError as e:
        assert str(e) == "Value value2 already exists for key1", "Unexpected ValueError"

    del d["key1"]
    assert "key1" in d, "deleting will keep last value if refcount>0"
    assert d["key1"] == "value1"

    del d["key1"]
    assert "key1" not in d, "deleting last reference deletes key"



from plexfuse.normalize import normalize


class DirEntry:
    def __init__(self, entries):
        self.entries = [normalize(str(s)) for s in entries]

    def __len__(self):
        return len(self.entries)

    def __iter__(self):
        return iter(self.entries)



from __future__ import annotations

from functools import cached_property
from pathlib import Path, PureWindowsPath
from typing import TYPE_CHECKING

from plexfuse.normalize import normalize

if TYPE_CHECKING:
    from plexapi.media import Guid, MediaPart, SubtitleStream
    from plexapi.video import Episode, Movie


class Playable:
    SUBTITLE_EXT = ("srt", "vtt")

    def __init__(self, item: Episode | Movie):
        self.item = item

    def __str__(self):
        return self.title

    @property
    def timestamps(self):
        return {
            "st_ctime": self.ctime,
            "st_atime": self.atime,
            "st_mtime": self.mtime,
        }

    @property
    def ctime(self):
        return self.item.addedAt.timestamp()

    @property
    def atime(self):
        if self.item.lastViewedAt is None:
            return self.mtime
        return self.item.lastViewedAt.timestamp()

    @property
    def mtime(self):
        if self.item.updatedAt is None:
            return self.ctime
        return self.item.updatedAt.timestamp()

    @property
    def guids(self) -> list[Guid]:
        return self.item.guids if self.item.guid.startswith("plex://") else []

    @cached_property
    def title(self):
        title = f"{self.item.seasonEpisode} " if self.item.TYPE == "episode" else ""
        title += self.item.title
        edition_title = self.item.__dict__.get("editionTitle", None)
        if edition_title:
            title += f" {{{edition_title}}}"

        year = self.item.__dict__.get("year", None)
        if year:
            title += f" ({year})"

        return normalize(title)

    @cached_property
    def subtitles(self):
        # Use basename of movie file if there's only one part
        basename = Path(next(iter(self.media_parts))).stem if len(self.media_parts) == 1 else None

        def inner():
            s: SubtitleStream
            for s in self.item.subtitleStreams():
                if s.key is None:
                    # Not downloadable. Embedded in video
                    continue
                if s.codec not in self.SUBTITLE_EXT:
                    print(f"Unsupported subtitle codec: {s.codec}: {self.item}")
                    # Not sure how to serve VOBSUB .idx files alone, where are .sub files?
                    continue
                language = s.languageCode or s.languageTag
                if basename:
                    title = f"{basename}.{language}.{s.codec}"
                else:
                    title = f"{s.language} ({language}).{s.codec}"
                yield title, s

        return dict(inner())

    @cached_property
    def media_parts(self):
        def inner():
            part: MediaPart
            for part in self.item.iterParts():
                # Remove directory part (Windows server on Unix)
                # We need to handle Windows and Unix differences,
                # hence the PureWindowsPath class
                yield PureWindowsPath(part.file).name, part

        return dict(inner())



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from plexfuse.normalize import normalize
from plexfuse.vfs.entry.EpisodeEntry import EpisodeEntry
from plexfuse.vfs.entry.MovieEntry import MovieEntry
from plexfuse.vfs.entry.SeasonEntry import SeasonEntry

if TYPE_CHECKING:
    from plexfuse.plex.types import SectionTypes


class SectionEntry:
    def __init__(self, section: SectionTypes):
        self.section = section
        self.title: str = normalize(section.title)
        self.type: str = section.type

    @cached_property
    def items(self):
        return [MovieEntry(m) for m in self.section.search()]

    @cached_property
    def items_by_title(self):
        return dict({(m.title, m) for m in self.items})

    @cached_property
    def seasons(self):
        return [SeasonEntry(s) for s in self.section.search(libtype="season")]

    @cached_property
    def episodes(self):
        return [EpisodeEntry(m) for m in self.section.search(libtype="episode")]

    def __str__(self):
        return self.title

    def __repr__(self):
        return str(self)



from __future__ import annotations

from functools import cached_property


class CachedPropertyCacheControl:
    """
    Cache control methods for an class using @cached_property
    """

    def __init__(self, obj):
        self.data = obj.__dict__
        self.name = obj.__class__.__name__
        self.members = self.find_members(obj.__class__)

    @staticmethod
    def find_members(cls):
        return [m for m in dir(cls) if isinstance(getattr(cls, m), cached_property)]

    def cache_info(self):
        yield f"{self.name} cache:"
        for k in self.members:
            v = self.data.get(k, None)
            f = f"{len(v)} items" if isinstance(v, (list, dict)) else v
            yield f"- {k}: {f}"

    def cache_clear(self):
        for k in self.members:
            try:
                del self.data[k]
                yield f"Deleted {self.name}: {k}"
            except KeyError:
                pass



import sys

if len(__package__) == 0:

    print(
        f"""

The '__main__' module does not seem to have been run in the context of a
runnable package ... did you forget to add the '-m' flag?

Usage: {sys.executable} -m plexfuse {' '.join(sys.argv[1:])}

"""
    )
    sys.exit(2)

from plexfuse.fs.main import main

# Ensure that program shows in usage (not __main__.py)
sys.argv[0] = __package__

main()



from __future__ import annotations

from typing import TYPE_CHECKING

from plexfuse.vfs.Playable import Playable

if TYPE_CHECKING:
    from plexapi.video import Episode


class EpisodeEntry(Playable):
    def __init__(self, item: Episode):
        super().__init__(item)



from __future__ import annotations

import os
from functools import cached_property
from os import makedirs
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING

from plexapi.exceptions import BadRequest, NotFound
from plexapi.server import PlexServer

from plexfuse.cache.HttpCache import HttpCache
from plexfuse.vfs.entry.LibraryEntry import LibraryEntry
from plexfuse.vfs.entry.MovieEntry import MovieEntry
from plexfuse.vfs.entry.SectionEntry import SectionEntry

if TYPE_CHECKING:
    from plexapi.video import Show


class PlexApi:
    CACHE_PATH = Path("cache")
    CACHE_VERSION = str(2)
    CHUNK_SIZE = 1024 * 1024 * 16
    HTTP_CACHE = False

    @cached_property
    def server(self):
        session = None
        if self.HTTP_CACHE:
            http_cache = HttpCache(self.CACHE_PATH / "http_cache")
            session = http_cache.session

        return PlexServer(session=session)

    @property
    def session(self):
        return self.server._session

    @property
    def token(self):
        return self.server._token

    @cached_property
    def library(self):
        return LibraryEntry(self.server.library)

    @property
    def section_types(self) -> list[str]:
        return list(self.library.section_types)

    def sections_by_type(self, type: str) -> list[SectionEntry] | None:
        try:
            return self.library.sections_by_type[type]
        except KeyError:
            return None

    def section_by_title(self, title: str) -> SectionEntry | None:
        try:
            return self.library.section_by_title[title]
        except KeyError:
            return None

    def library_items(self, library: str):
        try:
            return self.library.section_by_title[library].items
        except KeyError:
            return None

    def library_item(self, library: str, title: str):
        try:
            return self.library.section_by_title[library].items_by_title[title]
        except KeyError:
            return None

    def all_seasons(self, library):
        try:
            return self.library.section_by_title[library].seasons
        except KeyError:
            return None

    def all_episodes(self, library):
        try:
            return self.library.section_by_title[library].episodes
        except KeyError:
            return None

    def show_seasons(self, library: str, title: str):
        show: Show = self.library_item(library, title)
        if not show:
            return None
        result = [season for season in self.all_seasons(library)
                  if season.item.parentRatingKey == show.item.ratingKey]
        result.append(".plexmatch")
        return result

    def season_episodes(self, library: str, show_title: str, season_name: str):
        try:
            rating_key = [m.item.ratingKey
                          for m in self.show_seasons(library, show_title)
                          if m.title == season_name][0]
        except IndexError:
            return None

        return [m for m in self.all_episodes(library) if m.item.parentRatingKey == rating_key]

    def movie_files(self, library: str, title: str):
        movie = self.library_item(library, title)
        if movie is None:
            return None
        files = self.media_part_names(movie)
        subs = movie.subtitles
        if subs:
            files.extend(subs.keys())
        return files

    def episode_files(self, library: str, show_title: str, season_name: str, episode_title: str):
        episode = self.show_episode(library, show_title, season_name, episode_title)
        if not episode:
            return None
        files = self.media_part_names(episode)
        subs = episode.subtitles
        if subs:
            files.extend(subs.keys())
        return files

    def movie_part(self, library: str, title: str, part_name: str):
        movie = self.library_item(library, title)
        if movie is None:
            return None

        part = self.media_parts_by_name(movie, part_name)
        if part is None:
            return None

        return part, movie

    def episode_part(self, library: str, show_title: str, season_name: str, episode_title: str, part_name: str):
        episode = self.show_episode(library, show_title, season_name, episode_title)
        if not episode:
            return None
        part = self.media_parts_by_name(episode, part_name)
        return part, episode

    def show_episode(self, library: str, show_title: str, season_name: str, episode_title: str):
        show: Show = self.library_item(library, show_title)
        if not show:
            return None

        episodes = self.season_episodes(library, show_title, season_name)
        try:
            return [episode for episode in episodes if episode.title == episode_title][0]
        except IndexError:
            return None

    def media_part_names(self, media: MovieEntry):
        return list(self._media_part_names(media))

    @staticmethod
    def _media_part_names(media: MovieEntry):
        yield from media.media_parts.keys()
        if media.item.type == "movie":
            yield ".plexmatch"

    @staticmethod
    def media_parts_by_name(media: MovieEntry, filename: str):
        try:
            return media.media_parts[filename]
        except KeyError:
            return None

    def url(self, key: str, include_token=False):
        return self.server.url(key, includeToken=include_token)

    def fetch_item(self, key: str):
        return self.library.library.fetchItem(key)

    def cache_path(self, path: str):
        """Return cache path of item consisting machine uuid and a path"""
        return \
            self.CACHE_PATH \
            / self.server.machineIdentifier \
            / self.CACHE_VERSION \
            / path.lstrip("/")

    def request_file(self, key: str, size=None, offset=None):
        url = self.url(key)
        headers = {"X-Plex-Token": self.token}
        accepted_status = (200, 201, 204)
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Range
        if offset is not None:
            if size is not None:
                headers["Range"] = f"bytes={offset}-{offset + size - 1}"
            else:
                headers["Range"] = f"bytes={offset}-"
            accepted_status = (206,)
        elif size is not None:
            raise BadRequest("size not supported without offset")

        response = self.session.get(url, headers=headers, stream=True)
        if response.status_code not in accepted_status:
            if response.status_code == 404:
                raise NotFound(response.url)
            raise BadRequest(f"({response.status_code}): {response.url}")

        yield from response.iter_content(chunk_size=self.CHUNK_SIZE)

    def download_part(self, key: str, savepath: Path,
                      size=None, offset=None, overwrite=False):
        if overwrite is False and savepath.exists():
            return savepath

        content = self.request_file(key, size, offset)
        makedirs(savepath.parent, exist_ok=True)
        with NamedTemporaryFile(dir=savepath.parent, prefix=f"{savepath.stem}.") as handle:
            for chunk in content:
                handle.write(chunk)
            os.link(handle.name, savepath)

        return savepath



from __future__ import annotations

from typing import TYPE_CHECKING

from plexfuse.vfs.entry.AttrEntry import AttrEntry

if TYPE_CHECKING:
    from plexapi.media import MediaPart

    from plexfuse.vfs.ChunkedFile import ChunkedFile
    from plexfuse.vfs.Playable import Playable


class FileEntry(AttrEntry):
    def __init__(self, part: MediaPart, playable: Playable, reader: ChunkedFile):
        self.size = self.get_size(playable, part)
        self.key = part.key
        self.playable = playable
        self.reader = reader

    @property
    def attr(self):
        try:
            return self.playable.timestamps
        except AttributeError:
            return {}

    @staticmethod
    def get_size(playable: Playable, part: MediaPart):
        if playable.item.isPartialObject():
            playable.item.reload()
        part = next(partx for partx in playable.item.iterParts() if partx.key == part.key)
        size = part.size if part.exists else None
        return size

    def read(self, offset: int, size: int):
        return self.reader.read(self.key, size=size, offset=offset, max_size=self.size)

    def __len__(self):
        return self.size



#
# A FUSE filesystem for mounting a Plex Media Server in Python
#
# Mount with:
# $ plexfuse <mountpoint>
# And un-mount with:
# $ fusermount -u <mountpoint>
#

import fuse

from plexfuse.__version__ import __version__
from plexfuse.fs.PlexFS import PlexFS
from plexfuse.plex.PlexApi import PlexApi

fuse.fuse_python_api = (0, 2)


def main():
    usage = "PlexFS: A filesystem for mounting a Plex Media Server media\n\n"
    server = PlexFS(version=f"%prog: {__version__}\npython-fuse: {fuse.__version__}",
                    usage=usage + fuse.Fuse.fusage, dash_s_do="setsingle")
    server.parser.add_option(mountopt="cache_path", metavar="PATH",
                             default=PlexApi.CACHE_PATH,
                             help="cache downloads under PATH [default: %default]")
    server.parser.add_option(mountopt="http_cache",
                             default=False, action="store_true",
                             help="cache http requests using requests-cache")
    server.parse(values=server, errex=1)
    server.main()


if __name__ == "__main__":
    main()



class CacheControl:
    """
    Cache control methods for an object
    """

    def __init__(self, obj):
        self.name = obj.__class__.__name__
        self.methods = self.find_methods(obj)

    @staticmethod
    def find_methods(obj):
        methods = (getattr(obj, k) for k in dir(obj))
        return [m for m in methods if getattr(m, "cache_clear", None)]

    def cache_info(self):
        for m in self.methods:
            yield f"{self.name} cache: {m.__name__}: {m.cache_info()}"

    def cache_clear(self):
        for m in self.methods:
            size = m.cache_info().currsize
            m.cache_clear()
            yield f"{self.name} cache: {m.__name__}: cleared {size} entries"



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from plexfuse.cache.CacheControl import CacheControl
from plexfuse.cache.CachedPropertyCacheControl import \
    CachedPropertyCacheControl
from plexfuse.cache.DelayedPropertyCacheControl import \
    DelayedPropertyCacheControl
from plexfuse.cache.UserDictCacheControl import UserDictCacheControl
from plexfuse.vfs.entry.ControlEntry import ControlEntry
from plexfuse.vfs.entry.DirEntry import DirEntry

if TYPE_CHECKING:
    from plexfuse.fs.PlexFS import PlexFS
    from plexfuse.plex.PlexApi import PlexApi
    from plexfuse.vfs.PlexVFS import PlexVFS


class Control:
    def __init__(self, plex: PlexApi, plexfs: PlexFS, plexvfs: PlexVFS):
        self.plex = plex
        self.plexfs = plexfs
        self.plexvfs = plexvfs

    @property
    def root(self):
        return ["control"]

    @property
    def commands(self):
        return ["reload", "status"]

    @cached_property
    def cc_plexfs(self):
        return CacheControl(self.plexfs)

    @cached_property
    def cc_plexvfs(self):
        return UserDictCacheControl(self.plexvfs)

    @cached_property
    def cc_library(self):
        return CachedPropertyCacheControl(self.plex.library)

    @cached_property
    def cc_sections(self):
        return DelayedPropertyCacheControl(self.plex.library, "sections", CachedPropertyCacheControl)

    def reload(self):
        yield from self.cc_plexfs.cache_clear()
        yield from self.cc_plexvfs.cache_clear()
        yield from self.cc_sections.cache_clear()
        yield from self.cc_library.cache_clear()

    def status(self):
        yield from self.cc_plexfs.cache_info()
        yield from self.cc_plexvfs.cache_info()
        yield from self.cc_sections.cache_info()
        yield from self.cc_library.cache_info()

    def handle(self, pc: int, pe: list[str]):
        if pc == 1 and pe[0] in self.root:
            return DirEntry(self.commands)

        if pc != 2 or pe[0] != self.root[0]:
            return

        if pe[1] in self.commands:
            return ControlEntry(pe[1], getattr(self, pe[1]))



import errno
from functools import cache
from pathlib import Path
from threading import Lock

import fuse

from plexfuse.fs.PlexDirectory import PlexDirectory
from plexfuse.fs.PlexFile import PlexFile
from plexfuse.fs.RefCountedDict import RefCountedDict
from plexfuse.normalize import normalize
from plexfuse.plex.PlexApi import PlexApi
from plexfuse.vfs.entry.DirEntry import DirEntry
from plexfuse.vfs.PlexVFS import PlexVFS


class PlexFS(fuse.Fuse):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        plex = PlexApi()
        self.vfs = PlexVFS(plex, self)
        self.cache_path = None
        self.http_cache = None
        self.file_map = RefCountedDict()
        self.iolock = Lock()

    def fsinit(self):
        # "cache_path" property doesn't get always initialized from options:
        # https://github.com/libfuse/python-fuse/issues/61#issuecomment-1902472620
        cache_path = self.cache_path if self.cache_path else PlexApi.CACHE_PATH
        PlexApi.CACHE_PATH = Path(cache_path).absolute()
        PlexApi.HTTP_CACHE = self.http_cache
        print(f"fsinit: CACHE_PATH={PlexApi.CACHE_PATH}")
        print(f"fsinit: HTTP_CACHE={PlexApi.HTTP_CACHE}")

    @cache
    def getattr(self, path: str):
        path = normalize(path, is_path=True)
        try:
            item = self.vfs[path]
        except KeyError as e:
            print(f"ERROR: getattr: Unsupported path: {e}")
            return -errno.ENOENT

        if isinstance(item, DirEntry):
            return PlexDirectory(st_nlink=2 + len(item))

        return PlexFile(st_size=item.size, **item.attr)

    @cache
    def readdir(self, path: str, offset: int):
        return list(self._readdir(path))

    def _readdir(self, path: str):
        for r in [".", ".."]:
            yield fuse.Direntry(r)

        try:
            it = self.vfs[path]
        except KeyError as e:
            print(f"ERROR: readdir({path}): {e}")
            return

        for r in it:
            yield fuse.Direntry(r)

    def read(self, path, size, offset):
        with self.iolock:
            entry = self.file_map[path]

            return entry.read(offset, size)

    def release(self, path, flags):
        with self.iolock:
            try:
                del self.file_map[path]
            except KeyError as e:
                print(f"ERROR: release({path}): {e}")
                return -errno.EINVAL

            return 0

    def open(self, path, flags):
        with self.iolock:
            try:
                entry = self.vfs[path]
            except KeyError as e:
                print(f"ERROR: open({path}): {e}")
                return -errno.ENOENT

            if isinstance(entry, DirEntry):
                return -errno.EISDIR

            self.file_map[path] = entry

            return 0



from __future__ import annotations

import logging
from datetime import timedelta
from functools import cached_property
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


class HttpCache:
    def __init__(self, cache_path: Path):
        self.cache_path = cache_path

    @property
    def urls_expire_after(self):
        from requests_cache import DO_NOT_CACHE

        CACHE_1h = 60 * 60
        CACHE_1d = timedelta(days=1)

        return {
            "*/library/sections/*/all?includeGuids=1": CACHE_1d,
            "*/library/sections/*/all?includeGuids=1&type=3": CACHE_1h,
            "*/library/sections/*/all?includeGuids=1&type=4": CACHE_1h,
            # Item details
            "*/library/metadata/*?checkFiles=1*": CACHE_1d,
            # Some reload
            "*/library/metadata/*": CACHE_1d,
            # File download
            "*/library/parts/*": DO_NOT_CACHE,
            # Subtitles
            "*/library/streams/*": DO_NOT_CACHE,
            # default policy is not to cache
            "*": DO_NOT_CACHE,
        }

    @cached_property
    def session(self):
        from requests_cache import CachedSession

        # Set log levels
        loggers = [
            "requests_cache.session",
            "requests_cache.policy.actions",
            "requests_cache.policy.expiration",
            "requests_cache.backends",
            "requests_cache.backends.base",
            "requests_cache.backends.sqlite",
        ]
        stderr_handler = logging.StreamHandler()

        for name in loggers:
            logger = logging.getLogger(name)
            logger.setLevel(logging.DEBUG)
            logger.addHandler(stderr_handler)

        return CachedSession(
            cache_name=str(self.cache_path),
            # Plex sends "Cache-Control: no-cache" headers
            cache_control=False,
            urls_expire_after=self.urls_expire_after,
            # Plex doesn't Send Vary: X-Plex-Container-Start
            match_headers=["X-Plex-Container-Start"],
        )



import logging

from plexfuse.plex.PlexApi import PlexApi
from plexfuse.vfs.entry.MovieEntry import MovieEntry

logger = logging.getLogger(__name__)


def test_file():
    plex = PlexApi()
    path = "/library/metadata/1"
    filename = "1.webm"

    item = plex.fetch_item(path)
    media = MovieEntry(item)
    part = plex.media_parts_by_name(media, filename)
    if part:
        savepath = plex.cache_path(part.key)
        logger.error(savepath)
        plex.download_part(part.key, savepath)



import stat

from fuse import Stat

from plexfuse.fs.Timestampable import Timestampable


class PlexFile(Stat, Timestampable):
    def __init__(self, **kw):
        super().__init__(**kw)
        Timestampable.__init__(self)
        if self.st_mode is None:
            self.st_mode = stat.S_IFREG | 0o644
        self.st_nlink = 1



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from plexfuse.vfs.entry.AttrEntry import AttrEntry

if TYPE_CHECKING:
    from plexapi.media import Guid

    from plexfuse.vfs.Playable import Playable


class PlexMatchEntry(AttrEntry):
    def __init__(self, playable: Playable):
        self.playable = playable

    def read(self, offset: int, size: int):
        return self.content[offset:offset + size].encode()

    @property
    def attr(self):
        return self.playable.timestamps

    @cached_property
    def content(self):
        """
        Handler for .plexmatch files
        - https://support.plex.tv/articles/plexmatch/
        """
        return "\n".join(self.mapping(self.playable.guids))

    @staticmethod
    def mapping(guids: list[Guid]):
        for guid in guids:
            provider, value = guid.id.split("://")
            yield f"{provider}id: {value}"

        # Add newline to end of the file
        yield ""

    @cached_property
    def size(self):
        return len(self.content)

    def __len__(self):
        return self.size



from plexfuse.plex.PlexApi import PlexApi
from plexfuse.vfs.ChunkedFile import ChunkedFile


def test_file():
    plex = PlexApi()
    cf = ChunkedFile(plex)
    cf.chunk_size = 16

    assert cf.chunk_number(0) == 0
    assert cf.chunk_number(1) == 0
    assert cf.chunk_number(16) == 1
    assert cf.chunk_number(17) == 1
    assert cf.chunk_number(31) == 1

    assert cf.base_offset(0) == 0
    assert cf.base_offset(1) == 16
    assert cf.base_offset(2) == 32

    assert cf.chunk_offset(15) == 15
    assert cf.chunk_offset(16) == 0
    assert cf.chunk_offset(17) == 1
    assert cf.chunk_offset(31) == 15
    assert cf.chunk_offset(32) == 0
